package ru.dbotthepony.mc.otm.android.feature

import com.mojang.blaze3d.systems.RenderSystem
import com.mojang.blaze3d.vertex.PoseStack
import net.minecraft.ChatFormatting
import net.minecraft.network.FriendlyByteBuf
import net.minecraft.resources.ResourceLocation
import net.minecraft.server.level.ServerPlayer
import net.minecraft.sounds.SoundSource
import net.minecraftforge.network.NetworkEvent
import ru.dbotthepony.mc.otm.config.ClientConfig
import ru.dbotthepony.mc.otm.OverdriveThatMatters
import ru.dbotthepony.mc.otm.config.AndroidConfig
import ru.dbotthepony.mc.otm.android.AndroidResearchManager
import ru.dbotthepony.mc.otm.android.AndroidSwitchableFeature
import ru.dbotthepony.mc.otm.capability.MatteryPlayerCapability
import ru.dbotthepony.mc.otm.capability.energy.extractEnergyExact
import ru.dbotthepony.mc.otm.capability.matteryPlayer
import ru.dbotthepony.mc.otm.client.render.ResearchIcons
import ru.dbotthepony.mc.otm.core.math.Vector
import ru.dbotthepony.mc.otm.core.util.formatPower
import ru.dbotthepony.mc.otm.core.math.plus
import ru.dbotthepony.mc.otm.network.MatteryPacket
import ru.dbotthepony.mc.otm.network.MatteryPlayerNetworkChannel
import ru.dbotthepony.mc.otm.network.enqueueWork
import ru.dbotthepony.mc.otm.network.packetHandled
import ru.dbotthepony.mc.otm.network.sender
import ru.dbotthepony.mc.otm.registry.AndroidFeatures
import ru.dbotthepony.mc.otm.registry.MNames
import ru.dbotthepony.mc.otm.registry.MSoundEvents
import java.util.function.Supplier

object TriggerJumpBoostPacket : MatteryPacket {
	override fun write(buff: FriendlyByteBuf) {
		// no op
	}

	override fun play(context: Supplier<NetworkEvent.Context>) {
		context.packetHandled = true

		context.enqueueWork {
			val mattery = context.sender?.matteryPlayer ?: return@enqueueWork

			if (!mattery.isAndroid)
				return@enqueueWork

			val feature = mattery.getFeature(AndroidFeatures.JUMP_BOOST) as JumpBoostFeature? ?: return@enqueueWork

			if (feature.isActive && feature.cooldown <= 4 && mattery.androidEnergy.extractEnergyExact(AndroidConfig.JumpBoost.ENERGY_COST, false)) {
				feature.putOnCooldown()

				context.sender?.let {
					it.level.playSound(
						it,
						it,
						MSoundEvents.ANDROID_JUMP_BOOST,
						SoundSource.PLAYERS,
						1f,
						1f
					)
				}
			}
		}
	}
}

class JumpBoostFeature(capability: MatteryPlayerCapability) : AndroidSwitchableFeature(AndroidFeatures.JUMP_BOOST, capability) {
	private var tickCooldownClient = false

	override val maxCooldown: Int
		get() = (AndroidConfig.JumpBoost.BASE_COOLDOWN - AndroidConfig.JumpBoost.COOLDOWN_REDUCTION * level).coerceAtLeast(0)

	override var cooldown by synchronizer.int(setter = setter@{ value, access, setByRemote ->
		access.write(value)

		if (setByRemote) {
			tickCooldownClient = false
		}
	})

	private var lastGround = false

	fun movementTick(isJumping: Boolean, isShifting: Boolean) {
		check(ply !is ServerPlayer) { "Invalid side" }

		if (ply.isSpectator) {
			return
		}

		val old = lastGround
		lastGround = ply.isOnGround

		if (isActive && cooldown <= 0 && old != lastGround && !lastGround && isJumping && isShifting && ply.xRot <= ClientConfig.JUMP_BOOST_LOOK_ANGLE && android.androidEnergy.extractEnergyExact(AndroidConfig.JumpBoost.ENERGY_COST, true)) {
			ply.deltaMovement += Vector(0.0, AndroidConfig.JumpBoost.POWER * (level + 1) / 20.0, 0.0)
			putOnCooldown()
			MatteryPlayerNetworkChannel.sendToServer(TriggerJumpBoostPacket)

			ply.level.playSound(
				ply,
				ply,
				MSoundEvents.ANDROID_JUMP_BOOST,
				SoundSource.PLAYERS,
				1f,
				1f
			)
		}
	}

	override fun tickClient() {
		if (tickCooldownClient && cooldown > 0) {
			cooldown--
		}
	}

	override fun renderIcon(stack: PoseStack, x: Float, y: Float, width: Float, height: Float) {
		if (cooldown > 0) {
			RenderSystem.setShaderColor(1f, 0.4f, 0.4f, 1f)
		}

		ResearchIcons.ICON_JUMP_BOOST.render(stack, x, y, width, height)

		if (cooldown > 0) {
			RenderSystem.setShaderColor(1f, 1f, 1f, 1f)
		}
	}


	companion object {
		val POWER_COST_DESCRIPTION =
			AndroidResearchManager.descriptionFunc(
				ResourceLocation(OverdriveThatMatters.MOD_ID, MNames.JUMP_BOOST),
				"otm.gui.power_cost_per_use",
				{ AndroidConfig.JumpBoost.ENERGY_COST.formatPower().copy().withStyle(ChatFormatting.YELLOW) })
	}
}
