package ru.dbotthepony.mc.otm.block.entity.matter

import net.minecraft.core.BlockPos
import net.minecraft.server.level.ServerLevel
import net.minecraft.world.entity.player.Inventory
import net.minecraft.world.entity.player.Player
import net.minecraft.world.inventory.AbstractContainerMenu
import net.minecraft.world.item.ItemStack
import net.minecraft.world.level.Level
import net.minecraft.world.level.block.state.BlockState
import ru.dbotthepony.mc.otm.block.entity.MatteryDeviceBlockEntity
import ru.dbotthepony.mc.otm.block.entity.tech.BatteryBankBlockEntity
import ru.dbotthepony.mc.otm.capability.FlowDirection
import ru.dbotthepony.mc.otm.capability.MatteryCapability
import ru.dbotthepony.mc.otm.capability.matter.IMatterStorage
import ru.dbotthepony.mc.otm.container.MatteryContainer
import ru.dbotthepony.mc.otm.core.immutableList
import ru.dbotthepony.mc.otm.core.math.Decimal
import ru.dbotthepony.mc.otm.core.ifPresentK
import ru.dbotthepony.mc.otm.graph.Graph6Node
import ru.dbotthepony.mc.otm.graph.matter.IMatterGraphNode
import ru.dbotthepony.mc.otm.graph.matter.MatterNetworkGraph
import ru.dbotthepony.mc.otm.menu.matter.MatterCapacitorBankMenu
import ru.dbotthepony.mc.otm.registry.MBlockEntities

class MatterCapacitorBankBlockEntity(p_155229_: BlockPos, p_155230_: BlockState) : MatteryDeviceBlockEntity(MBlockEntities.MATTER_CAPACITOR_BANK, p_155229_, p_155230_), IMatterGraphNode, IMatterStorage {
	var gaugeLevel by synchronizer.float()
		private set

	override val matterNode = Graph6Node<IMatterGraphNode>(this)

	override val canSetMatterLevel: Boolean
		get() = false

	override var storedMatter: Decimal
		get() {
			var summ = Decimal.ZERO

			for (stack in container)
				if (!stack.isEmpty)
					stack.getCapability(MatteryCapability.MATTER).ifPresentK {
						summ += it.storedMatter
					}

			return summ
		}
		set(value) {
			throw UnsupportedOperationException()
		}

	override val maxStoredMatter: Decimal
		get() {
		var summ = Decimal.ZERO

		for (stack in container)
			if (!stack.isEmpty)
				stack.getCapability(MatteryCapability.MATTER).ifPresentK {
					summ += it.maxStoredMatter
				}

		return summ
	}

	override fun receiveMatter(howMuch: Decimal, simulate: Boolean): Decimal {
		if (!howMuch.isPositive)
			return Decimal.ZERO

		@Suppress("NAME_SHADOWING")
		var howMuch = howMuch
		var summ = Decimal.ZERO

		for (stack in container) {
			if (!stack.isEmpty) {
				stack.getCapability(MatteryCapability.MATTER).ifPresent {
					val diff = it.receiveMatter(howMuch, simulate)
					summ += diff
					howMuch -= diff
				}

				if (howMuch.isZero) {
					break
				}
			}
		}

		if (summ.isPositive && !simulate) {
			gaugeLevel = (storedMatter / maxStoredMatter).toFloat()
		}

		return summ
	}

	override fun extractMatter(howMuch: Decimal, simulate: Boolean): Decimal {
		if (!howMuch.isPositive)
			return Decimal.ZERO

		@Suppress("NAME_SHADOWING")
		var howMuch = howMuch
		var summ = Decimal.ZERO

		for (stack in container) {
			if (!stack.isEmpty) {
				stack.getCapability(MatteryCapability.MATTER).ifPresent {
					val diff = it.extractMatter(howMuch, simulate)
					summ += diff
					howMuch -= diff
				}

				if (howMuch.isZero) {
					break
				}
			}
		}

		if (summ.isPositive && !simulate) {
			gaugeLevel = (storedMatter / maxStoredMatter).toFloat()
		}

		return summ
	}

	override val matterFlow: FlowDirection
		get() = FlowDirection.BI_DIRECTIONAL

	val container = object : MatteryContainer(this::setChangedLight, BatteryBankBlockEntity.CAPACITY) {
		override fun setChanged(slot: Int, new: ItemStack, old: ItemStack) {
			super.setChanged(slot, new, old)
			capacitorStatus[slot].value = new.getCapability(MatteryCapability.MATTER).isPresent
			gaugeLevel = (storedMatter / maxStoredMatter).toFloat()
		}
	}.also(::addDroppableContainer)

	val capacitorStatus = immutableList(BatteryBankBlockEntity.CAPACITY) {
		synchronizer.bool(false)
	}

	init {
		savetable(::container, INVENTORY_KEY)
		exposeGlobally(MatteryCapability.MATTER, this)
		exposeGlobally(MatteryCapability.MATTER_NODE, this)
	}

	override fun createMenu(containerID: Int, inventory: Inventory, ply: Player): AbstractContainerMenu {
		return MatterCapacitorBankMenu(containerID, inventory, this)
	}

	override fun setRemoved() {
		super.setRemoved()
		matterNode.destroy(::MatterNetworkGraph)
	}

	override fun setLevel(level: Level) {
		super.setLevel(level)

		if (level is ServerLevel)
			MatterNetworkGraph.discoverFull(this, matterNode)
	}

	override fun getMatterHandler(): IMatterStorage {
		return this
	}
}
