@file:Suppress("FunctionName")

package ru.dbotthepony.mc.otm.client.screen.panels.slot

import com.mojang.blaze3d.systems.RenderSystem
import com.mojang.blaze3d.vertex.PoseStack
import net.minecraft.ChatFormatting
import net.minecraft.client.gui.GuiComponent
import net.minecraft.client.renderer.GameRenderer
import net.minecraft.world.inventory.AbstractContainerMenu
import net.minecraft.world.inventory.Slot
import net.minecraft.world.item.ItemStack
import ru.dbotthepony.mc.otm.client.minecraft
import ru.dbotthepony.mc.otm.client.render.MatterySprite
import ru.dbotthepony.mc.otm.client.render.Widgets18
import ru.dbotthepony.mc.otm.client.render.drawRect
import ru.dbotthepony.mc.otm.client.render.setDrawColor
import ru.dbotthepony.mc.otm.client.screen.MatteryScreen
import ru.dbotthepony.mc.otm.client.screen.panels.EditablePanel
import ru.dbotthepony.mc.otm.client.screen.panels.ISlotPanel
import javax.annotation.Nonnull
import kotlin.math.roundToInt

open class SlotPanel<out S : MatteryScreen<*>, out T : Slot> @JvmOverloads constructor(
	screen: S,
	parent: EditablePanel<*>?,
	final override val slot: T,
	x: Float = 0f,
	y: Float = 0f,
	width: Float = SIZE,
	height: Float = SIZE,
	noItemIcon: MatterySprite? = null
) : AbstractSlotPanel<S>(screen, parent, x, y, width, height, noItemIcon), ISlotPanel<T> {
	override fun mouseClickedInner(x: Double, y: Double, button: Int): Boolean {
		screen.returnSlot = slot
		return true
	}

	override fun mouseDraggedInner(x: Double, y: Double, button: Int, xDelta: Double, yDelta: Double): Boolean {
		screen.returnSlot = slot
		return true
	}

	override fun mouseReleasedInner(x: Double, y: Double, button: Int): Boolean {
		screen.returnSlot = slot
		return true
	}

	override val itemStack: ItemStack
		get() = slot.item

	override fun visibilityChanges(new: Boolean, old: Boolean) {
		if (!new) {
			slot.x = Int.MAX_VALUE
			slot.y = Int.MAX_VALUE
		}
	}

	override fun innerRender(stack: PoseStack, mouseX: Float, mouseY: Float, partialTick: Float) {
		slot.x = absoluteX.roundToInt() - screen.guiLeft
		slot.y = absoluteY.roundToInt() - screen.guiTop
		renderSlotBackground(stack, mouseX, mouseY, partialTick)

		var itemstack = slot.item
		val carried = screen.menu.carried
		var dragHit = false
		var countOverride: String? = null

		if (!carried.isEmpty && screen.isQuickCrafting && screen.quickCraftSlots.contains(slot)) {
			if (screen.quickCraftSlots.size == 1) {
				return
			}

			if (AbstractContainerMenu.canItemQuickReplace(slot, carried, true) && screen.menu.canDragTo(slot)) {
				itemstack = carried.copy()
				dragHit = true

				AbstractContainerMenu.getQuickCraftSlotCount(
					screen.quickCraftSlots,
					screen.quickCraftingType,
					itemstack,
					if (slot.item.isEmpty) 0 else slot.item.count
				)

				val k = Math.min(itemstack.maxStackSize, slot.getMaxStackSize(itemstack))
				if (itemstack.count > k) {
					countOverride = ChatFormatting.YELLOW.toString() + k
					itemstack.count = k
				}
			} else {
				screen.quickCraftSlots.remove(slot)
				screen.recalculateQuickCraftRemaining()
			}
		}

		RenderSystem.setShader(GameRenderer::getPositionTexShader)

		if (dragHit) {
			setDrawColor(SLOT_HIGHLIGHT_DRAG)
			drawRect(stack, 1f, 1f, width - 1, height - 1)
		}

		if (itemstack.isEmpty) {
			val icon = slot.noItemIcon

			if (icon != null) {
				val texture = minecraft.getTextureAtlas(icon.first).apply(icon.second)
				RenderSystem.setShaderTexture(0, texture.atlasLocation())
				GuiComponent.blit(stack, 1, 1, 0, 16, 16, texture)
			} else {
				noItemIcon?.render(stack, width = width, height = height)
			}
		}

		renderRegular(stack, itemstack, countOverride)

		if (isHovered) {
			screen.hoveredSlot = slot
		}
	}

	override fun innerRenderTooltips(@Nonnull stack: PoseStack, mouseX: Float, mouseY: Float, partialTick: Float): Boolean {
		// no op, screen does it for us (completely)
		return false
	}
}

fun <S : MatteryScreen<*>, T : Slot> BatterySlotPanel(
	screen: S,
	parent: EditablePanel<*>?,
	slot: T,
	x: Float = 0f,
	y: Float = 0f,
	width: Float = AbstractSlotPanel.SIZE,
	height: Float = AbstractSlotPanel.SIZE,
) = SlotPanel(screen, parent, slot, x, y, width, height, Widgets18.BATTERY_SLOT_BACKGROUND)

fun <S : MatteryScreen<*>, T : Slot> EquipmentBatterySlotPanel(
	screen: S,
	parent: EditablePanel<*>?,
	slot: T,
	x: Float = 0f,
	y: Float = 0f,
	width: Float = AbstractSlotPanel.SIZE,
	height: Float = AbstractSlotPanel.SIZE,
) = SlotPanel(screen, parent, slot, x, y, width, height, Widgets18.EQUIPMENT_BATTERY_SLOT_BACKGROUND)

fun <S : MatteryScreen<*>, T : Slot> PatternSlotPanel(
	screen: S,
	parent: EditablePanel<*>?,
	slot: T,
	x: Float = 0f,
	y: Float = 0f,
	width: Float = AbstractSlotPanel.SIZE,
	height: Float = AbstractSlotPanel.SIZE,
) = SlotPanel(screen, parent, slot, x, y, width, height, Widgets18.PATTERN_SLOT_BACKGROUND)

fun <S : MatteryScreen<*>, T : Slot> MatterCapacitorSlotPanel(
	screen: S,
	parent: EditablePanel<*>?,
	slot: T,
	x: Float = 0f,
	y: Float = 0f,
	width: Float = AbstractSlotPanel.SIZE,
	height: Float = AbstractSlotPanel.SIZE,
) = SlotPanel(screen, parent, slot, x, y, width, height, Widgets18.MATTER_CAPACITOR_SLOT_BACKGROUND)
